const _ = require('lodash');

module.exports = RemoteServiceManager;
module.exports.RemoteServiceManager = RemoteServiceManager;


/**
 * A factory for `BIServiceSDK` instances
 * (see affiliated [bi-service-sdk](https://github.com/BohemiaInteractive/bi-service-sdk) plugin)
 * and {@tutorial 4.SDK-integration}
 *
 * @param {Object} services
 * @param {Object} services.&lt;serviceName&gt; - dynamic index
 * @param {Object} services.&lt;serviceName&gt;.&lt;appName&gt; - dynamic index
 * @param {String} services.&lt;serviceName&gt;.&lt;appName&gt;.npm - npm package name which exports object that inherits BIServiceSDK
 * @param {String} services.&lt;serviceName&gt;.&lt;appName&gt;.host
 * @param {Boolean} services.&lt;serviceName&gt;.&lt;appName&gt;.ssl
 *
 * @constructor
 */
function RemoteServiceManager(services) {
    this.options = services || {};
    this.services = {};

    /**
     * `BIServiceSDK` constructor
     * @name RemoteServiceManager#BIServiceSDK
     * @instance
     * @type {Function}
     */
    let sdk = require('bi-service-sdk');
    // sdk.BIServiceSDK interface is depracated and sdk.BIServiceSDKInterface
    // should be used instead
    this.BIServiceSDK = sdk.BIServiceSDKInterface || sdk.BIServiceSDK;

}


/**
 * registers `BIServiceSDK` instance
 *
 * @public
 * @param {String} key - format `<serviceName>:<appName>`
 * @param {BIServiceSDK} sdk
 *
 * @return {RemoteServiceManager} - self
 */
RemoteServiceManager.prototype.add = function(key, sdk) {
    let keySegments = key.split(':')
    ,   serviceName = keySegments[0]
    ,   appName     = keySegments[1];

    if (!(sdk instanceof this.BIServiceSDK)) {
        throw new Error('sdk must be instanceof BIServiceSDK');
    }

    if (keySegments.length !== 2) {
        throw new Error('The first argument must be in format: `<serviceName>:<appName>`');
    }

    if (!this.services[serviceName]) {
        this.services[serviceName] = {};
    }

    if (!this.services[serviceName][appName]) {
        this.services[serviceName][appName] = {};
    }

    this.services[serviceName][appName][sdk.version] = sdk;

    return this;
};


/**
 * @public
 * @param {String} key - format `<serviceName>:<appName>:<version>`
 *
 * @return {BIServiceSDK}
 */
RemoteServiceManager.prototype.get = function(key) {
    let remoteService = _.get(this.services, key.split(':'));

    if (!remoteService) {
        throw new Error(`${key} remote service not found`);
    }

    return remoteService;
};


/**
 * constructs specific `BIServiceSDK` from configuration provided to the constructor  
 * the specific `BIServiceSDK` is supposed to be auto generated by `bi-service-sdk` plugin
 * and published as a npm package
 *
 * @example
 * //config.json5:
 * {
 *    services: {
 *       user: {
 *         public: {
 *           ssl: true,
 *           host: "127.0.0.1:4001",
 *           npm: "bi-user-public-sdk"
 *         }
 *       }
 *    }
 * }
 *
 * //method usage
 * remoteServiceManager.buildRemoteService('user:public:v1.0');
 *
 * @public
 *
 * @param {String} key - format: `<serviceName>:<appName>:<version>`
 * @param {Object} [options] - `BIServiceSDK` constructor options
 *
 * @throws {Error}
 * @return {BIServiceSDK}
 */
RemoteServiceManager.prototype.buildRemoteService = function(key, options) {
    let keySegments = key.split(':');
    let serviceName = keySegments[0];
    let appName     = keySegments[1];
    let version     = keySegments[2];

    if (keySegments.length !== 3) {
        throw new Error('The first argument must be in format: `<serviceName>:<appName>:<version>`');
    }

    let confPath = `${serviceName}:${appName}`;
    let conf = _.get(this.options, [serviceName, appName]);

    if (!conf) {
        throw new Error(`Cant find config value of "${confPath}"`);
    }

    if (!_.isPlainObject(conf) || !conf.npm) {
        throw new Error(`"services:${confPath}:npm" config option not found. Cant connect remote service.`);
    }

    let sdkVersions = module.require(conf.npm);

    if (!sdkVersions.hasOwnProperty(version)) {
        throw new Error(`${confPath} service sdk does not have version: ${version}`);
    }

    let opt = _.cloneDeep(options || {});
    let protocol = conf.protocol || (conf.ssl ? 'https' : 'http');
    opt.baseURL =  protocol + '://' + conf.host;

    let sdk = new sdkVersions[version](opt);

    this.add(`${serviceName}:${appName}`, sdk);

    return sdk;
};
